/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: partition-utils.c,v 1.69 2009/08/03 08:15:52 zlb Exp $ */

#include "phg.h"
#include "partition-sfc.h"
#include "hamilton.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* load balance efficiency */
static const FLOAT SFC_LBE = 1.004;

/* ************************************************
 * Hilbert Order 
 * used by phgImport
 **************************************************/

/* ordering:
 * used for ordering Hilbert order and Morton order*/
static int
comp_key(const void *n1, const void *n2)
{
    SFC *m1, *m2;
    int flag = 0;

    m1 = (SFC *) n1;
    m2 = (SFC *) n2;
    if (m1->sfc > m2->sfc)
        flag = 1;
    else if (m1->sfc < m2->sfc)
        flag = -1;
    else 
         flag = 0;
    return flag;
}

/* generate Hilbert Order */
/* 0: success, or fail */
int
phgHilbertOrder(GRID *g)
{
    SFC *x;
    INT i;
    SFC_FLOAT xmax, xmin;
    SFC_FLOAT ymax, ymin;
    SFC_FLOAT zmax, zmin;
    SFC_FLOAT etx, ety, etz;
    FLOAT x0, y0, z0;
    static const FLOAT lam[] = {0.25, 0.25, 0.25, 0.25};
    SIMPLEX *e;

    assert(g != NULL);
    if (g->nroot <= 0) {
        phgWarning("no element!\n");
        return 0;
    }
    if (g->nroot <= 2) {
        return 0;
    }

    x = (SFC *)phgAlloc(g->nroot * sizeof(SFC));
    if (x == NULL && g->nroot > 0) {
        phgError(1, "memory allocation error (%s:%d)\n", __FILE__, __LINE__);
        return -1;
    }

    /* scan all elements 
     * calculate barycentric coordinate */
    for (i = 0; i < g->nroot; i++) {
        e = g->roots + i;
        x[i].index = e->index;
        phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
        x[i].co[0] = x0;          /* x */
        x[i].co[1] = y0;          /* y */
        x[i].co[2] = z0;          /* z */
    } /* for */
    
    /* boundary box */
    xmin = g->bbox[0][0];
    ymin = g->bbox[0][1];
    zmin = g->bbox[0][2];
    xmax = g->bbox[1][0];
    ymax = g->bbox[1][1];
    zmax = g->bbox[1][2];

    /* scale bounding box  
     * such that x, y, z in [0, 1)^3 */
    etx = xmax - xmin;
    ety = ymax - ymin;
    etz = zmax - zmin;

    if (etx == 0.) etx = 1.;
    if (etx < ety) etx = ety;
    if (etx < etz) etx = etz;
    etx =  1. / etx;
    ety = etx;
    etz = etx;

    for (i = 0; i < g->nroot; i++) {
        x[i].co[0] = (x[i].co[0] - xmin) * etx;
        x[i].co[1] = (x[i].co[1] - ymin) * ety;
        x[i].co[2] = (x[i].co[2] - zmin) * etz;
    }
    /* generate Hilbert SFC */
    if (!phgSFCInvHilbert3D(x, g->nroot)) {
        phgWarning("error in generating Hilbert SFC (%s:%d)\n",
						__FILE__, __LINE__);
        phgFree(x);
        return -1;
    }

    /* sort result */
    qsort(x, g->nroot, sizeof(SFC), comp_key);
    /* reorder the indices */
    {
        INT *aux;
        INT kk;

        aux = (INT *)phgAlloc(g->nroot * sizeof(INT));
        if (aux == NULL) {
            phgWarning("memory allocation error (%s:%d)\n", __FILE__, __LINE__);
            phgFree(x);
            return -1;
        }

        for (i = 0; i < g->nroot; i++) {
            aux[x[i].index] = i;
        }

        for (i = 0; i < g->nroot; i++) {
            e = g->roots + i;
            kk = e->index;
            e->index = aux[kk];
        }
        phgFree(aux);
    }

    phgFree(x);
    return 0;
}

/* generate Morton Order */
/* 0: success, or fail */
int
phgMortonOrder(GRID *g)
{
    SFC *x;
    INT i;
    SFC_FLOAT xmax, xmin;
    SFC_FLOAT ymax, ymin;
    SFC_FLOAT zmax, zmin;
    SFC_FLOAT etx, ety, etz;
    FLOAT x0, y0, z0;
    static const FLOAT lam[] = {0.25, 0.25, 0.25, 0.25};
    SIMPLEX *e;

    assert(g != NULL);
    if (g->nroot <= 0) {
        phgWarning("no element!\n");
        return 0;
    }
    if (g->nroot <= 2) {
        return 0;
    }

    x = (SFC *)phgAlloc(g->nroot * sizeof(SFC));
    if (x == NULL && g->nroot > 0) {
        phgError(1, "memory allocation error (%s:%d)\n", __FILE__, __LINE__);
        return -1;
    }

    /* scan all elements 
     * calculate barycentric coordinate */
    for (i = 0; i < g->nroot; i++) {
        e = g->roots + i;
        x[i].index = e->index;
        phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
        x[i].co[0] = x0;          /* x */
        x[i].co[1] = y0;          /* y */
        x[i].co[2] = z0;          /* z */
    } /* for */
    
    /* boundary box */
    xmin = g->bbox[0][0];
    ymin = g->bbox[0][1];
    zmin = g->bbox[0][2];
    xmax = g->bbox[1][0];
    ymax = g->bbox[1][1];
    zmax = g->bbox[1][2];

    /* scale bounding box 
     * such that x, y, z in [0, 1)^3 */
    etx = xmax - xmin;
    ety = ymax - ymin;
    etz = zmax - zmin;

    if (etx == 0.) etx = 1.;
    if (etx < ety) etx = ety;
    if (etx < etz) etx = etz;
    etx =  1. / etx;
    ety = etx;
    etz = etx;

    for (i = 0; i < g->nroot; i++) {
        x[i].co[0] = (x[i].co[0] - xmin) * etx;
        x[i].co[1] = (x[i].co[1] - ymin) * ety;
        x[i].co[2] = (x[i].co[2] - zmin) * etz;
    }
    /* generate Morton SFC */
    if (!phgSFCInvMorton3D(x, g->nroot)) {
        phgWarning("error in generating Morton SFC (%s:%d)\n",
						__FILE__, __LINE__);
        phgFree(x);
        return -1;
    }
    /* sort result */
    qsort(x, g->nroot, sizeof(SFC), comp_key);
    /* reorder the indices */
    {
        INT *aux;
        INT kk;

        aux = (INT *)phgAlloc(g->nroot * sizeof(INT));
        if (aux == NULL) {
            phgWarning("memory allocation error (%s:%d)\n", __FILE__, __LINE__);
            phgFree(x);
            return -1;
        }

        for (i = 0; i < g->nroot; i++) {
            aux[x[i].index] = i;
        }

        for (i = 0; i < g->nroot; i++) {
            e = g->roots + i;
            kk = e->index;
            e->index = aux[kk];
        }
        phgFree(aux);
    }

    phgFree(x);
    return 0;
}

/* construct initial grid order 
 * default: hilbert order */
static const char *orders[] = {"hilbert", "morton", "hamilton", "none", NULL};
static int GridOrder = 0;

int
phgGridInitOrder(GRID *g, BOOLEAN dist)
{
    int ret = 1;
    double t1[3], t2[3];
    static BOOLEAN initialized = FALSE;

    /* register parameters */
    if (!initialized) {
        initialized = TRUE;
        phgOptionsRegisterTitle("RTK options:", NULL, "partition");
        phgOptionsRegisterKeyword("-rtk_init_order", "Initial ordering",
                orders, &GridOrder);
        return TRUE;
    }

    assert(g != NULL);
    if (dist) {  /* parallel version */
        phgError(1, "Not implemented yet!\n");
    }
    else {       /* sequential version */
        if (phgVerbosity > 0) {
            phgPrintf("Initial ordering for RTK: ");
            phgGetTime(t1);
        }
        
        /* Hilbert order */
        if (GridOrder == 0) {
            if (phgVerbosity > 0) {
                phgPrintf("Hilbert space-filling curve.\n");
            }
            ret = phgHilbertOrder(g);
        } /* Morton order */
        else if (GridOrder == 1) {
            if (phgVerbosity > 0) {
                phgPrintf("Morton space-filling curve.\n");
            }
            ret = phgMortonOrder(g);
        }    /* Hamilton Path */
        else if (GridOrder == 2) {
            if (phgVerbosity > 0) {
                phgPrintf("\tHamilton path.\n");
            }
            ret = phgHamiltonPath(g);
        }
	else {
            if (phgVerbosity > 0) {
                phgPrintf("none.\n");
            }
	    ret = 0;
	}
        
        /* success */
        if (ret == 0) {
            if (phgVerbosity > 0) {
                phgGetTime(t2);
                phgPrintf("Finished computing initial ordering of elements: ");
                phgPrintf("%fs\n", t2[2] - t1[2]);
                phgPrintf("Total roots: %d\n", g->nroot);
            }
        }
        else { /* fail */
            if (phgVerbosity > 0) {
                phgPrintf("Error in computing initial ordering.");
            }
        }
    }
    return ret;
}

#if USE_MPI
/***************************************************
 *             1D partitioner 
 **************************************************/
typedef struct {
    SFC_FLOAT low;
    SFC_FLOAT high;
    int   pn;
}PTNS;

/* tolerance for load balance efficiency */
static FLOAT LB_TOL;

/* return the partition id that key belongs to */
static int bsearch_ptn(PTNS *ptn, int p, SFC_FLOAT key);

/* max loops */
static int  MAXLOOPS = 10;

static int  CUTS = 16;

void phgPartition1DOptionsSet()
{
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
        initialized = TRUE;
        LB_TOL = SFC_LBE;
        phgOptionsRegisterTitle("Options for 1D partitioner:", NULL, 
                "partition");
        phgOptionsRegisterFloat("-lbe_tol", "Tolerance of load balance "
		"efficiency (>= 1.)", &LB_TOL);
        phgOptionsRegisterInt("-1d_cut", "number of cuts of 1D partitioner:"
		"10 <= cut <= 30 ", &CUTS);

        return;
    }
}

/* cut [0, 1] to p parts, such that lbe <= LB_TOL 
 * dts: store key \in [0, 1]
 * lx: length of dts 
 * p: # of partitions */
FLOAT
phgPartition1DV1(DOTS *dts, INT lx, int p, MPI_Comm comm)
{
    int npct = (p - 1) * CUTS + 1;  /* total  cuts */
    PTNS  ptn[npct];

    SFC_FLOAT lbe;                  /* load balance efficiency */
    SFC_FLOAT DotsSum;              /* total sum */
    SFC_FLOAT psum[p];              /* partial sum, psum[i] = sf * i * sum */
    SFC_FLOAT sum[npct];
    SFC_FLOAT tsum[npct];
    SFC_FLOAT bbox[p][2];           /* bounding box, 0: low, 1: high*/

    INT i,j;
    INT ctn;                        /* if we start next loop */
    int rank;                       /* pid */
    int flag;                       /* iterative times */
    int map[npct];                  /* partition result */

    /* check parameters */
    if (LB_TOL < 1.) {
        phgWarning("LB_TOL too small, changed to %lg\n", (double)SFC_LBE);
        LB_TOL = SFC_LBE;
    }
    else if (LB_TOL > 1.3) {
        phgWarning("LB_TOL too large (expecting poor load balance)\n");
    }
    assert((CUTS >= 10) && (CUTS <= 30));

    assert(lx >= 0 && p >= 1);
    if (lx > 0) {
        assert(dts != NULL);
    }

    /* deal with special case */
    if (p == 1) {  /* only one partition */
        for (i = 0; i < lx; i++) {
            dts[i].pn = 0;
        }
        return 1.;
    }
    
    /* no dot, just return */
    j = lx;
    MPI_Allreduce(&j, &ctn, 1, PHG_MPI_INT, MPI_SUM, comm);
    if (ctn == 0) {
        return 1.;
    }

    /* total sum */
    lbe = 0.;
    for (i = 0; i < lx; i++) {
        lbe += dts[i].w;
    }
    MPI_Allreduce(&lbe, &DotsSum, 1, PHG_MPI_SFC_FLOAT, MPI_SUM, comm);

    /* special case */
    if (DotsSum == 0.) {
        for (i = 0; i < lx; i++)
            dts[i].pn = 0;
        return 1.;
    }
    else if (DotsSum < 0.) {
        phgError(1, "negative weights (%s:%d)\n", __FILE__, __LINE__);
    }

    /* initialize partial sum, ideal condition
     * initialize bounding boxs */
    lbe = 1.0 / p;
    ctn = p - 1;
    for (i = 0; i < ctn; i++) {
        psum[i] = DotsSum * (i + 1.) * lbe;
        bbox[i][0] = 0.;
        bbox[i][1] = 1.;
    }
    psum[ctn] = DotsSum;
    bbox[ctn][0] = 0.;
    bbox[ctn][1] = 1.;

    /* initial guess partitions */
    lbe = 1. / npct;
    for (i = 0; i < npct; i++) {
        ptn[i].low = i * lbe;
        ptn[i].high = (i + 1) * lbe;
        ptn[i].pn = i;
    }

    MPI_Comm_rank(comm, &rank);

    /* partitioning 
     * phgRank 0 generates results then broadcast */
    flag = 0;
    ctn = 1;
    while (flag < MAXLOOPS && ctn) {
        SFC_FLOAT mxm;
        SFC_FLOAT tmp[p];
        INT ne[npct];  /* record the number of elements in each interval*/
        INT ne1[npct]; 
        int ct[npct];  /* record if current interval is overflow interval */
                       /* 0: no; > 0, yes and number of overflow intervals */
        int p1 = p - 1;
        int kk = 0;

        /* sum of current partitions */
        for (i = 0; i < npct; i++) {
            tsum[i] = 0.;
            ne1[i] = 0;
        }
        for (i = 0; i < lx; i++) {
            ctn = bsearch_ptn(ptn, npct, dts[i].key);
            dts[i].pn = ctn;
            tsum[ctn] += dts[i].w;
            ne1[ctn] += 1;
        }

        /* calculate total weights of all intervals */
        MPI_Reduce(tsum, sum, npct, PHG_MPI_SFC_FLOAT, MPI_SUM, 0, comm);

        /* calculate the number of elements in each interval */
        MPI_Reduce(ne1, ne, npct, PHG_MPI_INT, MPI_SUM, 0, comm);

        /* rank 0 generate partitions 
         * then broadcast to all processes */
        if (rank == 0) {
            /* prefix sum */
            tsum[0] = sum[0];
            ct[0] = 0;
            for (i = 1; i < npct; i++) {
                tsum[i] = sum[i];
                sum[i] += sum[i - 1];
                ct[i] = 0;
            }

            /* update bounding box */
            j = 0;
            for (i = 0; i < p1; i++) {
                /* look for new bounding box position */
                while (sum[j] < psum[i]) {
                    j++;
                }
                /* update bounding box */
                if (ptn[j].high < bbox[i][1]) bbox[i][1] = ptn[j].high;
                if (ptn[j].low  > bbox[i][0]) bbox[i][0] = ptn[j].low;
                ct[j] += 1;

                tmp[i] = 0.;
            }
            tmp[p - 1] = 0.;

            /* calculate max sum of current partition
             * map map[npct] to p parts */
            j = 0;
            for (i = 0; i < npct && j < p; i++) {
                if (sum[i] < psum[j]) {
                    map[i] = j;
                }
                else {
                    map[i] = j;
                    j++;

                    /* check if each overflow invterval has only one
                     * element.
                     * kk = 0 if yes. kk = 1 if no */
                    if ((ne[i] > 1) && (j < p1)) {
                        kk = 1;
                    }
                }
            }
            /* special case that j >= p */
            for (j = i; j < npct; j++) {
                map[j] = p1;
            }

            /* calculate load balance efficiency */
            for (i = 0; i < npct; i++) {
                j = map[i];
                tmp[j] += tsum[i];
            }
            mxm = tmp[0];
            for (i = 1; i < p; i++) {
                if (mxm < tmp[i]) mxm = tmp[i];
            }
            lbe = mxm * p / DotsSum;

            /* terminates if procedure satisfies stop criteria.
             * two stop criterias:
             * (#if 1) the first one terminates if lbe is less than or
             * equals to LB_TOL or each overflow interval has
             * only one element.
             * (#else part)the second one just uses lbe */
#if 1
            if ((lbe <= LB_TOL) || (kk == 0)) {
                ctn = 0;
            }
            else {
                ctn = 1;
            }
#else
            if (lbe <= LB_TOL) {
                ctn = 0;
            }
            else {
                ctn = 1;
            }
#endif

            if (phgVerbosity > 2) {
                phgPrintf("phgPartition1DV1 - LIF: %f\n", lbe);
            }
        }

        /* synchronize status
         * 1: start next iteration, 0: stop */
        MPI_Bcast(&ctn, 1, PHG_MPI_INT, 0, comm);
       
        if (ctn) {
            ctn = npct - 1;

            /* generate new partitions
             * cut p - 1 bounding boxes to npct parts */
            if (rank == 0) {
                SFC_FLOAT len;
                INT k, j1, j2;
                INT ctk;
                
                i = 0;
                j = 0;
                k = 0;
                while (i < p1) {
                    while (ct[j] < 1) {
                        j++;
                    }

                    /* new partition 
                     * we use array sum[ctn] for temp variable 
                     * sum[ctn] stores internal partitioning points */
                    j2 = ct[j] * CUTS;
#if 0
                    if (j < ctn) {
                        len = 1. / j2;

                        ctk = i + ct[j] - 1;
                        for (j1 = 1; j1 <= j2; j1++) {
                            sum[k] = (bbox[i][0] * (j2 - j1) 
                                    + bbox[ctk][1] * j1) * len;
                            k++;
                        }
                    }
                    else {
                        len = 1. / (j2 + 1.);
                        ctk = i + ct[j] - 1;
                        for (j1 = 1; j1 <= j2; j1++) {
                            sum[k] = (bbox[i][0] * (j2 + 1 - j1) 
                                    + bbox[ctk][1] * j1) * len;
                            k++;
                        }

                    }
#else
                    len = 1. / j2;
                    ctk = i + ct[j] - 1;

                    for (j1 = 1; j1 <= j2; j1++) {
                        sum[k] = (bbox[i][0] * (j2 - j1) 
                                + bbox[ctk][1] * j1) * len;
                        k++;
                    }
#endif
                    i += ct[j];
                    j++;
                }
#if 0
                /* check */
                for (i = 1; i < ctn; i++) {
                    assert(sum[i - 1] <= sum[i]);
                }
#endif
            }

            /* rank 0 broadcasts new partitions.
             * each process initializes new partitions */
            MPI_Bcast(sum, ctn, PHG_MPI_SFC_FLOAT, 0, comm);

            ptn[0].low = 0.;
            ptn[0].high = sum[0];
            for (i = 1; i < ctn; i++) {
                ptn[i].high = sum[i];
                ptn[i].low = sum[i - 1];
            }
            ptn[ctn].high = 1.;
            ptn[ctn].low = sum[ctn - 1];
            
            ctn = 1;
        }
        else {
            break;
        }

        flag++;
    }  /* while */

    /* synchronize results */
    MPI_Bcast(map, npct, PHG_MPI_INT, 0, comm);
    MPI_Bcast(&lbe, 1, PHG_MPI_SFC_FLOAT, 0, comm);

    for (i = 0; i < lx; i++) {
        j = dts[i].pn;
        dts[i].pn = map[j];
    }

    if (phgVerbosity > 0) {
        phgPrintf("phgPartition1DV1 - iterations: %d\n", flag);
        phgPrintf("phgPartition1DV1 - LIF: %f\n", lbe);
    }

    return lbe;
}

FLOAT
phgPartition1DV2(DOTS *dts, INT lx, int p, MPI_Comm comm)
{
    int npct = (p - 1) * CUTS + 1;  /* total  cuts */
    PTNS  ptn[npct];

    SFC_FLOAT lbe;                  /* load balance efficiency */
    SFC_FLOAT DotsSum;              /* total sum */
    SFC_FLOAT psum[p];              /* partial sum, psum[i] = sf * i * sum */
    SFC_FLOAT sum[npct];
    SFC_FLOAT tsum[npct];
    SFC_FLOAT bbox[p][2];           /* bounding box, 0: low, 1: high*/

    INT i,j;
    INT ctn;                        /* if we start next loop */
    int rank;                       /* pid */
    int flag;                       /* iterative times */
    int map[npct];                  /* partition result */
    int unpt;

    /* check parameters */
    if (LB_TOL < 1.) {
        phgWarning("LB_TOL too small, changed to %lg\n", (double)SFC_LBE);
        LB_TOL = SFC_LBE;
    }
    else if (LB_TOL > 1.3) {
        phgWarning("LB_TOL too large (expecting poor load balance)\n");
    }
    assert((CUTS >= 10) && (CUTS <= 30));

    assert(lx >= 0 && p >= 1);
    if (lx > 0) {
        assert(dts != NULL);
    }

    /* deal with special case */
    if (p == 1) {  /* only one partition */
        for (i = 0; i < lx; i++) {
            dts[i].pn = 0;
        }
        return 1.;
    }
    
    /* no dot, just return */
    j = lx;
    MPI_Allreduce(&j, &ctn, 1, PHG_MPI_INT, MPI_SUM, comm);
    if (ctn == 0) {
        return 1.;
    }

    /* total sum */
    lbe = 0.;
    for (i = 0; i < lx; i++) {
        lbe += dts[i].w;
    }
    MPI_Allreduce(&lbe, &DotsSum, 1, PHG_MPI_SFC_FLOAT, MPI_SUM, comm);

    /* special case */
    if (DotsSum == 0.) {
        for (i = 0; i < lx; i++)
            dts[i].pn = 0;
        return 1.;
    }
    else if (DotsSum < 0.) {
        phgError(1, "negative weights (%s:%d)\n", __FILE__, __LINE__);
    }

    /* initialize partial sum, ideal condition
     * initialize bounding boxs */
    lbe = 1.0 / p;
    ctn = p - 1;
    for (i = 0; i < ctn; i++) {
        psum[i] = DotsSum * (i + 1.) * lbe;
        bbox[i][0] = 0.;
        bbox[i][1] = 1.;
    }
    psum[ctn] = DotsSum;
    bbox[ctn][0] = 0.;
    bbox[ctn][1] = 1.;

    /* initial guess partitions */
    lbe = 1. / npct;
    for (i = 0; i < npct; i++) {
        ptn[i].low = i * lbe;
        ptn[i].high = (i + 1) * lbe;
        ptn[i].pn = i;
    }

    DotsSum = p / DotsSum;

    MPI_Comm_rank(comm, &rank);

    /* partitioning 
     * phgRank 0 generates results then broadcast */
    flag = 0;
    ctn = 1;
    unpt = npct;
    while (flag < MAXLOOPS && ctn) {
        SFC_FLOAT mxm;
        SFC_FLOAT tmp[p];
        INT ne[npct];  /* record the number of elements in each interval*/
        INT ne1[npct]; 
        int ct[npct];  /* record if current interval is overflow interval */
                       /* 0: no; > 0, yes and number of overflow intervals */
        int p1 = p - 1;
        int kk = 0;
        int skc;

        /* sum of current partitions */
        for (i = 0; i < npct; i++) {
            tsum[i] = 0.;
            ne1[i] = 0;
        }
        for (i = 0; i < lx; i++) {
            ctn = bsearch_ptn(ptn, unpt, dts[i].key);
            dts[i].pn = ctn;
            tsum[ctn] += dts[i].w;
            ne1[ctn] += 1;
        }

        /* calculate total weights of all intervals */
        MPI_Reduce(tsum, sum, npct, PHG_MPI_SFC_FLOAT, MPI_SUM, 0, comm);

        /* calculate the number of elements in each interval */
        MPI_Reduce(ne1, ne, npct, PHG_MPI_INT, MPI_SUM, 0, comm);

        /* rank 0 generate partitions 
         * then broadcast to all processes */
        if (rank == 0) {
            /* prefix sum */
            tsum[0] = sum[0];
            ct[0] = 0;
            for (i = 1; i < npct; i++) {
                tsum[i] = sum[i];
                sum[i] += sum[i - 1];
                ct[i] = 0;
            }

            /* update bounding box */
            j = 0;
            for (i = 0; i < p1; i++) {
                /* look for new bounding box position */
                while (sum[j] < psum[i]) {
                    j++;
                }
                /* update bounding box */
                if (ptn[j].high < bbox[i][1]) bbox[i][1] = ptn[j].high;
                if (ptn[j].low  > bbox[i][0]) bbox[i][0] = ptn[j].low;
                ct[j] += 1;

                tmp[i] = 0.;
            }
            tmp[p - 1] = 0.;

            /* calculate max sum of current partition
             * map map[npct] to p parts */
            j = 0;
            for (i = 0; i < npct && j < p; i++) {
                if (sum[i] < psum[j]) {
                    map[i] = j;
                }
                else {
                    map[i] = j;
                    j++;

                    /* check if each overflow invterval has only one
                     * element.
                     * kk = 0 if yes. kk = 1 if no */
                    if ((ne[i] > 1) && (j < p1)) {
                        kk = 1;
                    }
                }
            }
            /* special case that j >= p */
            for (j = i; j < npct; j++) {
                map[j] = p1;
            }

            /* calculate load balance efficiency */
            for (i = 0; i < npct; i++) {
                j = map[i];
                tmp[j] += tsum[i];
            }
            mxm = tmp[0];
            for (i = 1; i < p; i++) {
                if (mxm < tmp[i]) mxm = tmp[i];
            }
            lbe = mxm * DotsSum;

            /* terminates if procedure satisfies stop criteria.
             * two stop criterias:
             * (#if 1) the first one terminates if lbe is less than or
             * equals to LB_TOL or each overflow interval has
             * only one element.
             * (#else part)the second one just uses lbe */
#if 1
            if ((lbe <= LB_TOL) || (kk == 0)) {
                ctn = 0;
            }
            else {
                ctn = 1;
            }
#else
            if (lbe <= LB_TOL) {
                ctn = 0;
            }
            else {
                ctn = 1;
            }
#endif

            if (phgVerbosity > 2) {
                phgPrintf("phgPartition1DV2 - LIF: %f\n", lbe);
            }
        }

        /* synchronize status
         * 1: start next iteration, 0: stop */
        MPI_Bcast(&ctn, 1, PHG_MPI_INT, 0, comm);
       
        /* calculate the number of elements in each bounding box 
         * if all the numbers are 1, then terminate */
        if (ctn) {
            /* generate new partitions
             * cut p - 1 bounding boxes to npct parts */
            ctn = npct - 1;
            if (rank == 0) {
                SFC_FLOAT len;
                SFC_FLOAT md;    /* mid point of some partition */
                INT k, j1, j2;
                INT ctk;
#if 0
                SFC_FLOAT lbe1, lbe2;

                kk = 0;
                for (i = 1; i < p1; i++) {
                    if (bbox[i - 1][1] > bbox[i][0]) kk++;
                }

                i = 0;
                j = 0;
                k = 0;
                skc = 0;
                while (i < p1) {
                    while (ct[j] < 1) {
                        j++;
                    }

                    if (ct[j] == 1) {
                        lbe1 = sum[j] * DotsSum / (i + 1.);
                        lbe2 = tmp[i] * DotsSum;
                    }
                    else {
                        lbe1 = LB_TOL + 0.1;
                        lbe2 = LB_TOL + 0.1;
                    }
                    /* new partition 
                     * we use array tsum[ctn] for temp variable 
                     * tsum[ctn] stores internal partitioning points */
                    md = (bbox[i][0] + bbox[i][1]) * 0.5;

                    if ( ((lbe1 <= LB_TOL) && (lbe2 <= LB_TOL) && (lbe2 >= 1.0))
                      || ((ct[j] == 1) && (kk == 0) && (ne[j] <= 1)) 
                      || (md == bbox[i][0] || bbox[i][1] == md)){
                        tsum[k] = bbox[i][0];
                        k += 1;
                        tsum[k] = md;
                        k += 1;
                        tsum[k] = bbox[i][1];
                        k += 1;

                        i += ct[j];
                        j++;
                        skc++;
                    }
                    else {
                        j2 = ct[j] * CUTS;
                        len = 1. / j2;
                        ctk = i + ct[j] - 1;

                        for (j1 = 1; j1 <= j2; j1++) {
                            tsum[k] = (bbox[i][0] * (j2 - j1) 
                                    + bbox[ctk][1] * j1) * len;
                            k++;
                        }
                        i += ct[j];
                        j++;
                    }
                }
#else
                kk = 0;
                for (i = 1; i < p1; i++) {
                    if (bbox[i - 1][1] > bbox[i][0]) kk++;
                }

                i = 0;
                j = 0;
                k = 0;
                skc = 0;
                while (i < p1) {
                    while (ct[j] < 1) {
                        j++;
                    }

                    /* new partition 
                     * we use array tsum[ctn] for temp variable 
                     * tsum[ctn] stores internal partitioning points */
                    md = (bbox[i][0] + bbox[i][1]) * 0.5;

                    if (((ct[j] == 1) && (kk == 0) && (ne[j] <= 1)) 
                      || (md == bbox[i][0] || bbox[i][1] == md)){
                        tsum[k] = bbox[i][0];
                        k += 1;
                        tsum[k] = md;
                        k += 1;
                        tsum[k] = bbox[i][1];
                        k += 1;

                        i += ct[j];
                        j++;
                        skc++;
                    }
                    else {
                        j2 = ct[j] * CUTS;
                        len = 1. / j2;
                        ctk = i + ct[j] - 1;

                        for (j1 = 1; j1 <= j2; j1++) {
                            tsum[k] = (bbox[i][0] * (j2 - j1) 
                                    + bbox[ctk][1] * j1) * len;
                            k++;
                        }
                        i += ct[j];
                        j++;
                    }
                }
#endif
                
                unpt = k + 1;
                for (i = k; i <= ctn; i++) {
                    tsum[i] = 1.;
                }
#if 0
                printf("skc: %d k: %d npct: %d \n", skc, k, npct);
                /* check */
                for (i = 1; i < ctn; i++) {
                    assert(tsum[i - 1] <= tsum[i]);
                }
#endif
            }

            /* rank 0 broadcasts new partitions.
             * each process initializes new partitions */
            MPI_Bcast(tsum, ctn, PHG_MPI_SFC_FLOAT, 0, comm);
            MPI_Bcast(&skc, 1, MPI_INT, 0, comm);
            MPI_Bcast(&unpt, 1, MPI_INT, 0, comm);

            if (skc < p1) {
                ptn[0].low = 0.;
                ptn[0].high = tsum[0];
                for (i = 1; i < ctn; i++) {
                    ptn[i].high = tsum[i];
                    ptn[i].low = tsum[i - 1];
                }
                ptn[ctn].high = 1.;
                ptn[ctn].low = tsum[ctn - 1];

                ctn = 1;
            }
            else {
                ctn = 0;
            }
        }
        else {
            break;
        }

        flag++;
    }  /* while */

    /* synchronize results */
    MPI_Bcast(map, npct, PHG_MPI_INT, 0, comm);
    MPI_Bcast(&lbe, 1, PHG_MPI_SFC_FLOAT, 0, comm);

    for (i = 0; i < lx; i++) {
        j = dts[i].pn;
        dts[i].pn = map[j];
    }

    if (phgVerbosity > 0) {
        phgPrintf("phgPartition1DV2 - iterations: %d\n", flag);
        phgPrintf("phgPartition1DV2 - LIF: %f\n", lbe);
    }

    return lbe;
}
static int
bsearch_ptn(PTNS *ptn, int p, SFC_FLOAT key)
{
    int lo, hi;
    int md;
    int ret;
    
    assert(p >= 1);
    
    lo = 0;
    hi = p - 1;
    md = (lo + hi) >> 1;
    while (1) {
        if (key >= ptn[md].high) ret = 1;
        else if (key - ptn[md].low < 0.) ret = -1;
        else ret = 0;

        if (ret == 1) {
            lo = md;
            md = (lo + hi) >> 1;
            if (md == lo) md = hi;
        }
        else if (ret == -1) {
            hi = md;
            md = (lo + hi) >> 1;
        }
        else {
            break;
        }
    }
    return md;
}

/* phgPartitionRemap, remap the partitions such that the
 * thhe mount of distributed data will be small
 * nprocs: partition number
 * return value > 0, if partitions change;
 * return value = 0, if partitions unchange;
 * return value < 0, if subroutines fail and then partitions keep the same.
 *
 * The algorithm comes from PLUM: 
 * parallel load balancing for adaptive unstructured meshed 
 * Leonid Oliker, Rupak Biswas 
 *
 * heuristic algorithm 
 * phgPartitionRemap stores the whole similiar matrix.
 * phgPartitionRemapV2 is a improved version which 
 * stores the non-zero entries only.
 * The second version can use memories much more efficient,
 * and will be more competitive when the number of processes
 * are large */

/* similiar matrix */
typedef struct {
    INT s;      /* number of elements */
    int pid;    /* current pid */
    int ptn;    /* partition id */
} SM;

/* descending order used by qsort */
static int
comp_sml(const void *m1, const void *m2)
{
    SM *n1, *n2;
    int flag = 0;
    
    n1 = (SM *)m1;
    n2 = (SM *)m2;

    if (n1->s > n2->s)
        flag = -1;
    else if (n1->s == n2->s)
        flag = 0;
    else if (n1->s < n2->s)
        flag = 1;
    return flag;
}

int
phgPartitionRemap(GRID *g, int nprocs, MPI_Comm comm)
{
    SM *local, *sml = NULL;
    SIMPLEX *e;
    MPI_Datatype type;
    MPI_Datatype types[3];
    MPI_Aint indices[3];
    INT i, j;
    int *result;
    int rank, size;
    int blocklens[3];
    int ret = 0;

    i = g->nleaf;
    MPI_Allreduce(&i, &j, 1, PHG_MPI_INT, MPI_SUM, comm);
    if ((j == 0) || (nprocs <= 1)) {
        return ret;
    }

    local = (SM *) phgAlloc(nprocs * sizeof(SM));
    result = (int *) phgAlloc(nprocs * sizeof(int));

    if ((local == NULL) || (result == NULL)) {
        phgError(1, "memory allocation error (%s:%d)\n", __FILE__, __LINE__);
    }

    /* create a new datatype for exchange information */
    /* used by structure tree */
    blocklens[0] = blocklens[1] = blocklens[2] = 1;
    types[0] = PHG_MPI_INT;
    types[1] = types[2] = MPI_INT;
    MPI_Address(&local[0].s, &indices[0]);
    MPI_Address(&local[0].pid, &indices[1]);
    MPI_Address(&local[0].ptn, &indices[2]);
    indices[2] -= indices[0];
    indices[1] -= indices[0];
    indices[0] = 0;
    MPI_Type_struct(3, blocklens, indices, types, &type);
    MPI_Type_commit(&type);

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    /* gather information to creat similiar matrix */
    for (i = 0; i < nprocs; i++) {
        local[i].s = 0;
        local[i].pid = rank;
        local[i].ptn = i;
    }
    ForAllElements(g, e) {
        local[e->mark].s += 1;
    }

    if (rank == 0) {
        /* allocate space for process 0 */
        sml = (SM *) phgAlloc(nprocs * nprocs * sizeof(SM));
        if (sml == NULL) {
	    phgError(1, "memory allocation error (%s:%d)\n", __FILE__,__LINE__);
        }
    }

    /* build similiar matrix */
    MPI_Gather(local, nprocs, type, sml, nprocs, type, 0, comm);
    MPI_Type_free(&type);
    phgFree(local);

    /* map algorithm by PLUM */
    if (rank == 0){
        int *part_map;
        int *proc_unmap;
        int *t1;
        int count;
        SM *p;

        part_map = (int *) phgAlloc(nprocs * sizeof(int));
        proc_unmap = (int *) phgAlloc(nprocs * sizeof(int));
        t1 = (int *) phgAlloc(nprocs * sizeof(int));

        if ((part_map == NULL) || (proc_unmap == NULL) || (t1 == NULL)) {
            phgError(1, "memory allocation error (%s:%d)\n", __FILE__,__LINE__);
        }

        /* inlarge similiar matrix */
        /* such that sml is square */
        if (size < nprocs) {
            for (i = size; i < nprocs; i++) {
                for (j = 0; j < nprocs; j++) {
                    count = i * nprocs + j;
                    sml[count].pid = i;
                    sml[count].ptn = j;
                    sml[count].s = 0;
                }
            }
        }

        /* initialize */
        for (i = 0; i < nprocs; i++) {
            proc_unmap[i] = 1;
            part_map[i] = 0;
        }
        /* order the similiar matrix, descending order */
        qsort(sml, size * nprocs, sizeof(SM), comp_sml);

        /* map algorithm */
        count = 0;
        p = sml;
        while (count < nprocs) {
            while ((proc_unmap[p->pid] == 0 || part_map[p->ptn] == 1)) {
                p++;
            }
            proc_unmap[p->pid] = 0;
            part_map[p->ptn] = 1;
            result[p->ptn] = p->pid;
            count++;
        } /* map algorithm */

        /* verification 
         * count = 0: valid result 
         * count > 0: invalid result.*/
        count = 0;
        for (i = 0; i < nprocs; i++) {
            t1[i] = 0;

            /* check if final partitions have proper indices.
             * count = 0 if yes, > 0 if not */
            if ((result[i] < 0) || (result[i] > nprocs)) {
                count++;
            }
        }

        /* check if each index was mapped just once 
         * count = 0 if yes, > 0 if not */
        if (count == 0) {
            for (i = 0; i < nprocs; i++) {
                t1[result[i]] += 1;
            }
            for (i = 0; i < nprocs; i++) {
                if (t1[i] != 1)
                    count++;
            }
        }

        if (count == 0) { /* succeed */
            /* default: partitions unchanged. */
            ret = 0;

            /* check if partitions are changed */
            for (i = 0; i < nprocs; i++) {
                if (result[i] != i) ret = 1;
            }
        }
        else { /* fail */
            ret = -1;
        }

        phgFree(part_map);
        phgFree(proc_unmap);
        phgFree(t1);
    }

    /* broadcast results.
     * if partitions are changed and then we would remap partitions */
    MPI_Bcast(&ret, 1, MPI_INT, 0, comm);
    if (ret > 0) {
        MPI_Bcast(result, nprocs, MPI_INT, 0, comm);
        ForAllElements(g, e) {
            i = e->mark;
            e->mark = result[i];
        }
        if (phgVerbosity > 2) {
            phgPrintf("phgPartitionRemap - map results\n");
            for (i = 0; i < nprocs; i++) {
                phgPrintf("map: %6d  ---> %6d\n", i, result[i]);
            }
        }
    }
    else {
        if (phgVerbosity > 0)
            phgPrintf("phgPartitionRemap - partitions unchanged\n");
    }

    phgFree(sml);
    phgFree(result);

    return ret;
}

int
phgPartitionRemapV2(GRID *g, int nprocs, MPI_Comm comm)
{
    SM *local, *sml = NULL;
    SIMPLEX *e;
    MPI_Datatype type;
    MPI_Datatype types[3];
    MPI_Aint indices[3];
    INT i;
    INT nt;
    int *result;
    int rank, size;
    int nonz;
    int blocklens[3];
    int len[nprocs];
    int ofst[nprocs];
    int ret = 0;

    i = g->nleaf;
    MPI_Allreduce(&i, &nt, 1, PHG_MPI_INT, MPI_SUM, comm);
    if ((nt == 0) || (nprocs <= 1)) {
        return ret;
    }

    local = (SM *) phgAlloc(nprocs * sizeof(SM));
    result = (int *) phgAlloc(nprocs * sizeof(int));

    if ((local == NULL) || (result == NULL)) {
        phgError(1, "memory allocation error (%s:%d)\n", __FILE__, __LINE__);
    }

    /* create a new datatype for exchange information */
    /* used by structure tree */
    blocklens[0] = blocklens[1] = blocklens[2] = 1;
    types[0] = PHG_MPI_INT;
    types[1] = types[2] = MPI_INT;
    MPI_Address(&local[0].s, &indices[0]);
    MPI_Address(&local[0].pid, &indices[1]);
    MPI_Address(&local[0].ptn, &indices[2]);
    indices[2] -= indices[0];
    indices[1] -= indices[0];
    indices[0] = 0;
    MPI_Type_struct(3, blocklens, indices, types, &type);
    MPI_Type_commit(&type);

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    /* gather information to creat similiar matrix */
    for (i = 0; i < nprocs; i++) {
        local[i].s = 0;
        local[i].pid = rank;
        local[i].ptn = i;
    }
    ForAllElements(g, e) {
        local[e->mark].s += 1;
    }

    /* calculate non-zero entries of each process.
     * and move non-zero entries to the front part of the array */
    nonz = 0;
    for (i = 0; i < nprocs; i++) {
        if (local[i].s > 0) {
            local[nonz].s = local[i].s;
            local[nonz].pid = local[i].pid;
            local[nonz].ptn = local[i].ptn;
            nonz++;
        }
    }
    /* process 0 gathers the number of non-zero entries from 
     * all processes and calculate offset  */
    MPI_Gather(&nonz, 1, MPI_INT, len, 1, MPI_INT, 0, comm);

    if (rank == 0) {
        ofst[0] = 0;
        for (i = 1; i < size; i++) {
            ofst[i] = ofst[i - 1] + len[i - 1];
        }

        /* allocate space for process 0 */
        nt = ofst[size - 1] + len[size - 1];
        sml = (SM *) phgAlloc(nt * sizeof(SM));
        if (sml == NULL) {
            phgError(1, "memory allocation error (%s:%d)\n", __FILE__,__LINE__);
        }
    }

    /* process 0 build similiar matrix */
    MPI_Gatherv(local, nonz, type, sml, len, ofst, type, 0, comm);
    MPI_Type_free(&type);
    phgFree(local);

    /* map algorithm by PLUM */
    if (rank == 0){
        int *part_map;
        int *proc_unmap;
        int *t1;
        int count;
        int kk;
        SM *p;

        part_map = (int *) phgAlloc(nprocs * sizeof(int));
        proc_unmap = (int *) phgAlloc(nprocs * sizeof(int));
        t1 = (int *) phgAlloc(nprocs * sizeof(int));

        if ((part_map == NULL) || (proc_unmap == NULL) || (t1 == NULL)) {
            phgError(1, "memory allocation error (%s:%d)\n", __FILE__,__LINE__);
        }

        /* initialize */
        for (i = 0; i < nprocs; i++) {
            proc_unmap[i] = 1;
            part_map[i] = 0;
            result[i] = -1;
        }
        /* order the similiar matrix, descending order */
        qsort(sml, nt, sizeof(SM), comp_sml);

        /* map algorithm */
        count = 0;
        kk = 0;
        p = sml;
        nt -= 1;
        while ((count < nprocs) && (kk < nt)) {
            while ( (proc_unmap[p->pid] == 0 || part_map[p->ptn] == 1)
                 && (kk < nt)) {
                p++;
                kk++;
            }
            if ((proc_unmap[p->pid] != 0) && (part_map[p->ptn] != 1)) {
                proc_unmap[p->pid] = 0;
                part_map[p->ptn] = 1;
                result[p->ptn] = p->pid;
                count++;
            }
        } /* map algorithm */

        count = 0;
        for (i = 0; i < nprocs; i++) {
            if (result[i] < 0) count++;
            t1[i] = 0;
        }

        /* deal with special case that some arn't mapped */
        if (count > 0) {
            for (i = 0; i < nprocs; i++) {
                if (result[i] >= 0) {
                    t1[result[i]] = -1;
                }
            }
            kk = 0;
            for (i = 0; i < nprocs; i++) {
                if (result[i] < 0) {
                    while (t1[kk] < 0) {
                        kk++;
                    }
                    result[i] = kk;
                    kk++;
                }
            }
        }

        /* verification */
        /* TRUE: valid result   */
        /* FALSE: invalid result. unmap. */
        count = 0;
        for (i = 0; i < nprocs; i++) {
            t1[i] = 0;

            /* check if final partitions have proper indices.
             * count = 0 if yes, > 0 if not */
            if ((result[i] < 0) || (result[i] > nprocs)) {
                count++;
            }
        }

        /* check if each index was mapped just once 
         * count = 0 if yes, > 0 if not */
        if (count == 0) {
            for (i = 0; i < nprocs; i++) {
                t1[result[i]] += 1;
            }
            for (i = 0; i < nprocs; i++) {
                if (t1[i] != 1)
                    count++;
            }
        }

        if (count == 0) { /* succeed */
            /* default: partitions unchanged. */
            ret = 0;

            /* check if partitions are changed */
            for (i = 0; i < nprocs; i++) {
                if (result[i] != i) ret = 1;
            }
        }
        else { /* fail */
            ret = -1;
        }

        phgFree(part_map);
        phgFree(proc_unmap);
        phgFree(t1);
    }

    /* broadcast results.
     * if partitions are changed and then we would remap partitions */
    MPI_Bcast(&ret, 1, MPI_INT, 0, comm);
    if (ret > 0) {
        MPI_Bcast(result, nprocs, MPI_INT, 0, comm);
        ForAllElements(g, e) {
            i = e->mark;
            e->mark = result[i];
        }
        if (phgVerbosity > 2) {
            phgPrintf("phgPartitionRemapV2 - map results\n");
            for (i = 0; i < nprocs; i++) {
                phgPrintf("map: %6d  ---> %6d\n", i, result[i]);
            }
        }
    }
    else {
	if (phgVerbosity > 0)
            phgPrintf("phgPartitionRemapV2: partitions unchanged\n\n");
    }

    phgFree(sml);
    phgFree(result);

    return ret;
}

/* print grid quality information */
/* this function must be called just after phgBalanceGrid */
/* or e->mark isn't changed.   */
void
phgPartitionQuality(GRID *g)
{
    INT i, k;
    INT cr, ci, cb;
    INT tcr, tci, tcb;
    INT en;
    FLOAT si, tsi;
    SIMPLEX *e;
    INT *pid;

    if (g->nroot <= 0)
        return;
    if (phgRank >= g->nprocs)
        return;

    en = ci = cb = cr = 0;
    ForAllElements(g, e) {
        en++;
        for (i = 0; i < NFace; i++) {
            if (e->bound_type[i] & REMOTE) {
                cr++;
                continue;
            }
            if (e->bound_type[i] & INTERIOR) {
                ci++;
                continue;
            }
            cb++;
        }
    }

    /* load imbalance efficiency */
    MPI_Allreduce(&en, &k, 1, PHG_MPI_INT, MPI_MAX, g->comm);
    si = (FLOAT)k * g->nprocs / (FLOAT)g->nelem_global;
    phgPrintf("g->lif:    %f\n", (double)si);

    /* total cut edges */
    MPI_Allreduce(&cr, &tcr, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    tcr >>= 1;
    phgPrintf("Cut edges: %d\n", tcr);

    /* surface indics */
    ci >>= 1;
    k = cb + ci + cr;
    /* local surface index */
    if (k > 0) {
        si = (FLOAT) cr / (FLOAT) k;
    }
    else {
        si = 0;
    }
    MPI_Allreduce(&si, &tsi, 1, PHG_MPI_FLOAT, MPI_MAX, g->comm);
    phgPrintf("Maximum local surface index: %f\n", (double)tsi);

    /* average surface index */
    MPI_Allreduce(&si, &tsi, 1, PHG_MPI_FLOAT, MPI_SUM, g->comm);
    si = tsi / (FLOAT)g->nprocs;
    phgPrintf("Average surface index:       %f\n", (double)si);

    /* global surface index */
    MPI_Allreduce(&cb, &tcb, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    MPI_Allreduce(&ci, &tci, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    k = tci + tcr + tcb;

    si = (FLOAT) tcr / (FLOAT)k;
    phgPrintf("Global surface index:        %f\n", (double)si);

    /* compute interprocess connectivity */
    pid = phgAlloc((g->nprocs + 1) * sizeof(*pid));
    for (i = 0; i < g->nprocs; i++) {
        if (i != phgRank && g->neighbours.counts[i] != 0) {
            pid[i] = 1;
        }
        else {
            pid[i] = 0;
        }
    }
    k = 0;
    for (i = 0; i < g->nprocs; i++)
        k += pid[i];
    phgFree(pid);
    MPI_Allreduce(&k, &i, 1, PHG_MPI_INT, MPI_MAX, g->comm);
    phgPrintf("Max interprocess connectivity: %d\n", i);
}
#endif
