/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg-mpi.c,v 1.39 2009/07/12 15:48:01 wleng Exp $ */

#include "phg.h"

#if USE_ZOLTAN
#include "partition-zoltan.h"
#endif	/* USE_ZOLTAN */

GRID *_phg_g_sys = NULL;

#if USE_MPI

#include <stdlib.h>

/* for getpid() */
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>	/* PATH_MAX */

int phgRank = 0, phgNProcs = 0, phgCmdSerialno = 0;

static int functions_count = 0;
static PFunc_t *functions = NULL;
static int mpi_initialized = FALSE;

#if DEBUG
char (*phgProcessorNames)[MPI_MAX_PROCESSOR_NAME + 1] = NULL;

int _phg_comm_count = 0, _phg_comm_count_max = 0;
#undef MPI_Comm_dup
#undef MPI_Comm_create
#undef MPI_Comm_free
int (*MPI_Comm_split_real)(MPI_Comm comm, int color, int key, MPI_Comm *commp);
int (*MPI_Comm_dup_real)(MPI_Comm comm, MPI_Comm *comm_out);
int (*MPI_Comm_create_real)(MPI_Comm comm, MPI_Group group, MPI_Comm *commp);
int (*MPI_Comm_free_real)(MPI_Comm *commp);
#endif	/* DEBUG */

int
phgRegisterParallelFunction(PFunc_t func, const char *funcname,
				const char *file, int line)
{
    if (phgInitialized) {
	phgInfo(-1, "%s:%d, unregistered parallel function called.\n",
		file, line);
	phgInfo(-1, "Please add the following line:\n");
	phgInfo(-1, "    RegisterParallelFunction(%s);\n", funcname);
	phgInfo(-1, "before phgInit(...).\n");
	phgError(1, "abort.\n");
    }

    phgInfo(3, "registering parallel function %d: %s\n",
	functions_count, funcname);
    functions = phgRealloc_(functions, sizeof(PFunc_t) * (functions_count + 1),
					sizeof(PFunc_t) * functions_count);
    functions[functions_count] = func;

    return functions_count++;
}

static void
finalize_mpi(GRID *g)
{
    Unused(g);

    ParallelFunction(finalize_mpi, TRUE);

    if (functions != NULL) {
	phgFree(functions);
	functions = NULL;
	functions_count = 0;
    }

    if (phgNProcs <= 0)
	return;

    if (!mpi_initialized) {
	MPI_Finalize();
    }
    else {
	/* to prevent slave processes from exiting */
	phgRank = 0;
    }

    if (phgMasterSlave && phgRank > 0)
	exit(0);

    phgNProcs = -1;

    return;
}

void
phgInitMPI(int *argc, char **argv[])
{
    char name[MPI_MAX_PROCESSOR_NAME + 1], cwd[PATH_MAX + 1];
    int len;

#if DEBUG
    MPI_Comm_split_real	= MPI_Comm_split;
    MPI_Comm_dup_real	= MPI_Comm_dup;
    MPI_Comm_create_real= MPI_Comm_create;
    MPI_Comm_free_real	= MPI_Comm_free;
#endif	/* DEBUG */

    MPI_Initialized(&mpi_initialized);
    /* save cwd */
    getcwd(cwd, sizeof(cwd));
    if (!mpi_initialized)
	if (MPI_Init(argc, argv))
	    phgError(1, "MPI_Init failed, abort.\n");
    /* broadcast and restore cwd */
    MPI_Bcast(cwd, sizeof(cwd), MPI_BYTE, 0, MPI_COMM_WORLD);
    chdir(cwd);

    MPI_Comm_rank(MPI_COMM_WORLD, &phgRank);
    MPI_Comm_size(MPI_COMM_WORLD, &phgNProcs);

    /* Initialize parallel functions */
    RegisterParallelFunction(finalize_mpi);
    RegisterParallelFunction(phgSetVerbosity_);
    RegisterParallelFunction(phgNewGrid_);
    RegisterParallelFunction(phgPartitionGrid__);
    RegisterParallelFunction(phgRedistributeGrid);
    RegisterParallelFunction(phgRefineMarkedElements);
    RegisterParallelFunction(phgCoarsenMarkedElements);
    RegisterParallelFunction(phgRefineAllElements_);
    RegisterParallelFunction(phgRefineRandom_);
    RegisterParallelFunction(phgRefineSurface);
    RegisterParallelFunction(phgCheckConformity);
    RegisterParallelFunction(phgFreeGrid_);
    RegisterParallelFunction(phgDumpGridInfo);
    RegisterParallelFunction(phgDumpGrid);
    RegisterParallelFunction(phgDumpTree);
    RegisterParallelFunction(phgExportVTK_);
    RegisterParallelFunction(phgExportDX_);
    RegisterParallelFunction(phgExportEnsight_);

    MPI_Get_processor_name(name, &len);
    name[len] = '\0';
#if DEBUG
    len = sizeof(phgProcessorNames[0]);
    phgProcessorNames = phgAlloc(phgNProcs * len);
    MPI_Allgather(name, len, MPI_BYTE, phgProcessorNames, len, MPI_BYTE,
		  MPI_COMM_WORLD);
#endif

    phgOptionsParseCmdline(argc, argv);

    /* Initialize modules after parsing all options.
     * Note: also add the calls below in phgInit() in utils.c */
    phgPerfInit();			/* init phgPerf */
    phgSolverInitialize(argc, argv);	/* init phgSolver */
#if USE_ZOLTAN
    phgPartitionZoltanInitialize(*argc, *argv);
#endif	/* USE_ZOLTAN */

    phgInitialized = TRUE;
    phgOptionsHelp();

    if (phgRank == 0) {
	phgPrintf("Parallel Hierarchical Grid (version " PHG_VERSION ").\n");
	fflush(stdout);
	if (phgVerbosity >= 2) {
	    phgPrintf("sizeof(SIMPLEX)    = %d\n", sizeof(SIMPLEX));
	    phgPrintf("sizeof(GRID)       = %d\n", sizeof(GRID));
	    phgPrintf("sizeof(RNEIGHBOUR) = %d\n", sizeof(RNEIGHBOUR));
	}
	if (phgVerbosity >= 1) {
	    phgPrintf("number of process(es): %d\n", phgNProcs);
	    phgPrintf("%s mode.\n", phgMasterSlave ? "Master-slave" : "SPMD");
	}
	fflush(stdout);
    }
    phgInfo(1, ">>> processor=%s, pid=%d <<<\n", name, getpid());

    phgTrapSignals_(FALSE);
    phgTimerInit(FALSE);

    if (phgRank > 0 && phgMasterSlave) {
	if (_phg_pause)
	    phgPause(0);
	phgSlave();
    }

    return;
}

void
phgFinalizeMPI(void)
{
    finalize_mpi(_phg_g_sys);

#if DEBUG
    phgFree(phgProcessorNames);
    phgProcessorNames = NULL;
#endif
}

void
phgCallSlaves(int cmd)
{
    if (phgMasterSlave) {
	phgInfo(3, "bringing slave to function %d\n", cmd);
	MPI_Bcast(&cmd, 1, MPI_INT, 0, MPI_COMM_WORLD);
	phgCmdSerialno++;
    }
}

void
phgSlave(void)
{
    int cmd;

    while (TRUE) {
	/* wait for next command */
	phgInfo(3, "phgSlave: waiting for command.\n");
	MPI_Bcast(&cmd, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if (cmd >= functions_count) 
	    phgError(1, "phgSlave: invalid command %d\n", cmd);
	phgCmdSerialno++;
	phgInfo(3, "entering function %d, g_sys=%p\n", cmd, _phg_g_sys);
	functions[cmd](_phg_g_sys);
	phgInfo(3, "returning from function %d, g_sys=%p\n", cmd, _phg_g_sys);
    }
}

#else	/* USE_MPI */

int phgRank = 0, phgNProcs = 1, phgCmdSerialno = 0;

#endif /* USE_MPI */
