/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-hamg.c,v 1.12 2009/05/28 01:37:30 zlb Exp $ */

/* Two-level + AMG solver for high-order FEM discretization of the Helmholtz
 * operator -div(a grad u) + b u, based on Hypre BoomerAMG
 * (note: only for HBn currently)
 *
 * Test:
 *	poisson -solver pcg -pcg_pc_type solver \
 *		-pcg_pc_opts "-solver hamg" -dof_type HB2
 *
 * WARNING: don't use this solver since the code is experimental!!!
 */

#include "phg.h"
#include "solver-phg.h"	/* phgSolverJacobi__() */

#include <string.h>	/* memcpy() */
#include <strings.h>	/* bzero() */

#if USE_HYPRE	/* whole file */

#define Initialize		NULL
#define Finalize		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

typedef struct {
   SOLVER	*hsolver;	/* high order solver */
   SOLVER	*lsolver;	/* low order solver */
} OEM_DATA;

#define	_t	((OEM_DATA *)solver->oem_data)

/* Note: for hierarchical basis the transfer operator and low order matrix
 * can be directly constructed from the high order matrix */
static MAT *
low_order_matrix(SOLVER *solver)
{
    GRID *g;
    MAP *map;
    MAT *mat;
    MAT_ROW *row;
    INT i, j, k, c, N0, N1, ncols, *cols;
    FLOAT *data;
    int rank;

    g = solver->mat->rmap->dofs[0]->g;
    map = phgMapCreateSimpleMap(g->comm, g->nvert_owned, g->nvert_global);
    mat = phgMapCreateMat(map, map);

    /* copy entries */
    ncols = 0;
    cols = NULL;
    data = NULL;
    N0 = solver->mat->rmap->partition[g->rank];
    N1 = solver->mat->rmap->partition[g->rank + 1];
    for (i = 0; i < g->nvert_owned; i++) {
	row = (void *)phgMatGetRow(solver->mat, i);
	if (row->ncols > ncols) {
	    ncols = row->ncols;
	    phgFree(cols);
	    cols = phgAlloc(ncols * sizeof(*cols));
	    phgFree(data);
	    data = phgAlloc(ncols * sizeof(*data));
	}
	for (j = k = 0; k < row->ncols; k++) {
	    if ((c = row->cols[k]) >= N0 && c < N1) {
		/* in-process column */
		if (c >= N0 + g->nvert_owned)
		    continue;
		c = c - N0 + map->partition[g->rank];
	    }
	    else {
		/* off-process column (FIXME: use binary search) */
		for (rank = 0; rank < g->nprocs; rank++)
		    if (c < solver->mat->rmap->partition[rank + 1])
			break;
		assert(rank < g->nprocs);
		c = c - solver->mat->rmap->partition[rank]
		      + map->partition[rank];
		if (c >= map->partition[rank + 1])
		    continue;
	    }
	    cols[j] = c;
	    data[j] = row->data[k];
	    j++;
	}
	phgMatAddLGEntries(mat, 1, &i, j, cols, data);
    }

    phgMapDestroy(&map);
    phgMatAssemble(mat);

    return mat;
}

static int
RegisterOptions(void)
{
#if 0
    phgOptionsRegisterTitle("\nHAMG options:", NULL, "hamg");
#endif
    return 0;
}

static int
Init(SOLVER *solver)
{
    solver->oem_data = phgCalloc(1, sizeof(OEM_DATA));

    return 0;
}

static int
Create(SOLVER *solver)
{
    phgOptionsPush();

    phgOptionsSetOptions("-solver hypre -solver_maxit 1 "
			 "-solver_atol 0. -solver_btol 0. -solver_rtol 0. "
			 "-hypre_solver boomeramg -hypre_pc none");

    /* low order AMG solver */
    assert(solver->mat->rmap->dofs != NULL);
    if (solver->mat->rmap->nglobal >
	solver->mat->rmap->dofs[0]->g->nvert_global) {
	MAT *mat = low_order_matrix(solver);
	if (mat != NULL) {
	    _t->lsolver = phgSolverMat2Solver(SOLVER_DEFAULT, mat);
	    _t->lsolver->rhs_updated = TRUE;
	    phgMatDestroy(&mat);
	}
    }

    /* high order smoother */
    /*phgOptionsSetOptions("-hypre_amg_max_levels 1");*/
    _t->hsolver = phgMat2Solver(SOLVER_DEFAULT, solver->mat);
    _t->hsolver->rhs_updated = TRUE;

    phgOptionsPop();

    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    phgSolverDestroy(&_t->hsolver);
    phgSolverDestroy(&_t->lsolver);
    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
Assemble(SOLVER *solver)
{
    if (solver->sent_to_oem == TRUE)
	return 0;
    solver->sent_to_oem = TRUE;

    return 0;
}
 
static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    VEC *y, *y0;
    FLOAT r0;
    INT i;

    Assemble(solver);

    r0 = -1.0;
    solver->nits = 0;
    y = phgVecCopy(x, NULL);
    phgVecCopy(solver->rhs, &_t->hsolver->rhs);
    if (_t->lsolver != NULL)
	y0 = phgMapCreateVec(_t->lsolver->rhs->map, 1);
    else
	y0 = NULL;

#if 1
/*-------------------------------------------------------------------*/
if (_t->lsolver != NULL) {
#if 1
	    phgVecCopy(solver->rhs, &x);

	    /* diagonal preconditioner */
	    if (solver->mat->diag == NULL)
		phgMatSetupDiagonal(solver->mat);
	    for (i = 0; i < x->map->nlocal; i++)
		x->data[i] /= solver->mat->diag[i];
#else
	    bzero(x->data, x->map->nlocal * sizeof(*x->data));
	    phgSolverJacobi__(solver->mat, solver->rhs, &x, 1, 0.,0.,0., NULL);
#endif

	    /* low order preconditioner */
	    phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &_t->hsolver->rhs);
	    memcpy(_t->lsolver->rhs->data, _t->hsolver->rhs->data, 
			y0->map->nlocal * sizeof(*y0->data));
	    bzero(y0->data, y0->map->nlocal * sizeof(*y0->data));
	    phgSolverVecSolve(_t->lsolver, FALSE, y0);
	    for (i = 0; i < y0->map->nlocal; i++)
		x->data[i] += y0->data[i];
	    return solver->nits = 1;
}
/*-------------------------------------------------------------------*/
#endif

    do {
//phgVecCopy(solver->rhs, &_t->hsolver->rhs);
//phgVecCopy(x, &y);
	phgMatVec(MAT_OP_N, -1.0, solver->mat, y, 1.0, &_t->hsolver->rhs);
	solver->residual = phgVecNorm2(_t->hsolver->rhs, 0, NULL);
	if (r0 < 0.0) {
	    r0 = solver->residual;
	    if (r0 == 0.0)
		break;
	    r0 = 1. / r0;
	    solver->residual = 1.0;
	}
	else {
	    solver->residual *= r0;
	}
	if (phgVerbosity > 0)
	    phgPrintf("\tnits = %d, residual = %le\n", solver->nits,
			(double)solver->residual);
	if (solver->residual <= solver->rtol || solver->nits >= solver->maxit)
	    break;

	bzero(y->data, y->map->nlocal * sizeof(*y->data));
	phgSolverVecSolve(_t->hsolver, FALSE, y);

	if (_t->lsolver != NULL) {
	    /* low order solver */
//phgVecCopy(solver->rhs, &_t->hsolver->rhs);
//phgVecCopy(x, &y);
	    phgMatVec(MAT_OP_N, -1.0, solver->mat, y, 1.0, &_t->hsolver->rhs);
	    memcpy(_t->lsolver->rhs->data, _t->hsolver->rhs->data, 
			y0->map->nlocal * sizeof(*y0->data));
	    bzero(y0->data, y0->map->nlocal * sizeof(*y0->data));
	    phgSolverVecSolve(_t->lsolver, FALSE, y0);
	    for (i = 0; i < y0->map->nlocal; i++)
		y->data[i] += y0->data[i];
	}

	phgVecAXPBY(1.0, y, 1.0, &x);

#if 0
phgVecCopy(solver->rhs, &_t->hsolver->rhs);
phgVecCopy(x, &y);
	phgMatVec(MAT_OP_N, -1.0, solver->mat, y, 1.0, &_t->hsolver->rhs);
	bzero(y->data, y->map->nlocal * sizeof(*y->data));
	phgSolverVecSolve(_t->hsolver, FALSE, y);
	phgVecAXPBY(1.0, y, 1.0, &x);
#endif

	++solver->nits;
    } while (TRUE);
    phgVecDestroy(&y);
    phgVecDestroy(&y0);

    if (destroy) {
	phgSolverDestroy(&_t->hsolver);
	phgSolverDestroy(&_t->lsolver);
	solver->oem_data = NULL;
    }

    return solver->nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverHAMG_ = {
    "HAMG", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy,
    AddMatrixEntries, AddRHSEntries,
    Assemble, Solve, NULL, NULL, NULL, TRUE, TRUE
};

#endif	/* USE_HYPRE */
